using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Kav
{
    public class DeferredPBREffect : Effect
    {
        EffectParameter gPositionParam;
        EffectParameter gAlbedoParam;
        EffectParameter gNormalParam;
        EffectParameter gMetallicRoughnessParam;
        EffectParameter shadowMapParam;

        EffectParameter directionalLightColorParam;
        EffectParameter directionalLightDirectionParam;

        EffectParameter eyePositionParam;
        EffectParameter lightSpaceMatrixParam;

        PointLightCollection pointLightCollection;

        public Texture2D GPosition { get; set; }
        public Texture2D GAlbedo { get; set; }
        public Texture2D GNormal { get; set; }
        public Texture2D GMetallicRoughness { get; set; }
        public Texture2D ShadowMap { get; set; }

        public Vector3 DirectionalLightColor { get; set; }
        public Vector3 DirectionalLightDirection { get; set; }
        public Matrix LightSpaceMatrix { get; set; }

        public Vector3 EyePosition { get; set; }

        public int MaxPointLights { get; } = 64;

        public PointLightCollection PointLights
        {
            get { return pointLightCollection; }
            private set { pointLightCollection = value; }
        }

        public DeferredPBREffect(GraphicsDevice graphicsDevice) : base(graphicsDevice, Resources.DeferredPBREffect)
        {
            CacheEffectParameters();

            pointLightCollection = new PointLightCollection(
                Parameters["PointLightPositions"],
                Parameters["PointLightColors"],
                MaxPointLights
            );
        }

        protected DeferredPBREffect(DeferredPBREffect cloneSource) : base(cloneSource)
        {
            GPosition = cloneSource.GPosition;
            GAlbedo = cloneSource.GAlbedo;
            GNormal = cloneSource.GNormal;
            GMetallicRoughness = cloneSource.GMetallicRoughness;

            EyePosition = cloneSource.EyePosition;

            PointLights = new PointLightCollection(
                Parameters["LightPositions"],
                Parameters["PositionLightColors"],
                MaxPointLights
            );

            for (int i = 0; i < MaxPointLights; i++)
            {
                PointLights[i] = cloneSource.PointLights[i];
            }
        }

        public override Effect Clone()
        {
            return new DeferredPBREffect(this);
        }

        protected override void OnApply()
        {
            gPositionParam.SetValue(GPosition);
            gAlbedoParam.SetValue(GAlbedo);
            gNormalParam.SetValue(GNormal);
            gMetallicRoughnessParam.SetValue(GMetallicRoughness);
            shadowMapParam.SetValue(ShadowMap);

            directionalLightColorParam.SetValue(DirectionalLightColor);
            directionalLightDirectionParam.SetValue(DirectionalLightDirection);

            eyePositionParam.SetValue(EyePosition);
            lightSpaceMatrixParam.SetValue(LightSpaceMatrix);
        }

        void CacheEffectParameters()
        {
            gPositionParam = Parameters["gPosition"];
            gAlbedoParam = Parameters["gAlbedo"];
            gNormalParam = Parameters["gNormal"];
            gMetallicRoughnessParam = Parameters["gMetallicRoughness"];
            shadowMapParam = Parameters["shadowMap"];

            directionalLightDirectionParam = Parameters["DirectionalLightDirection"];
            directionalLightColorParam = Parameters["DirectionalLightColor"];
            lightSpaceMatrixParam = Parameters["LightSpaceMatrix"];

            eyePositionParam = Parameters["EyePosition"];
        }
    }
}
